from pathlib import Path
from typing import Optional, Union, Sequence, Set
from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as _snapatac2

def call_peaks(
    data: Union[AnnData, AnnDataSet],
    group_by: Union[str, Sequence[str]],
    selections: Optional[Set[str]] = None,
    q_value: float = 0.05,
    out_dir: Optional[Path] = None,
    key_added: str = 'peaks',
):
    """
    Call peaks using MACS2.

    Use the `callpeak` command in MACS2 to identify regions enriched with TN5
    insertions. The parameters passed to MACS2 are:
    "-shift -100 -extsize 200 -nomodel -callsummits -nolambda -keep-dup all"

    The results are stored in `.uns[key_added]`.

    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    group_by
        Group the cells before peak calling. If a `str`, groups are obtained from
        `.obs[group_by]`.
    selections
        Call peaks for the selected groups only.
    q_value
        q_value cutoff used in MACS2.
    out_dir
        If provided, raw peak files from each group will be saved in the directory.
        Otherwise, they will be stored in a temporary directory which will be removed
        afterwards.
    key_added
        `.uns` key under which to add the peak information.
    """
    if isinstance(group_by, str):
        group_by = data.obs[group_by].astype("str")
    out_dir = out_dir if out_dir is None else str(out_dir)
    _snapatac2.call_peaks(data, group_by, selections, q_value, out_dir, key_added)


from typing import Optional, Union
from snapatac2._snapatac2 import AnnData
import numpy as np
from ._utils import render_plot

def umap(
    adata: AnnData,
    color: Union[str, np.ndarray],
    use_rep: Optional[str] = None,
    marker_size: float = 2,
    marker_opacity: float = 0.5,
    width: float = 550,
    height: float = 500,
    show: bool = True,
    interactive: bool = True,
    out_file: Optional[str] = None,
):
    """
    Plot UMAP embedding

    Parameters
    ----------
    adata
        Annotated data matrix.
    color
        If the input is a string, it will be used the key to retrieve values from
        `obs`.
    use_rep
    marker_size
    marker_opacity
    width
    height
    interactive
        Whether to make interactive plot
    out_file
        Path of the output file for saving the output image, end with '.svg' or '.pdf' or '.png'

    Returns
    -------
    
    """
    import plotly.express as px
    from natsort import index_natsorted
    import pandas as pd

    use_rep = "X_umap" if use_rep is None else use_rep
    embedding = adata.obsm[use_rep] 

    if isinstance(color, str):
        groups = adata.obs[color]
    else:
        groups = color

    idx = index_natsorted(groups)
    embedding = embedding[idx, :]
    groups = groups[idx]

    if embedding.shape[1] >= 3:
        df = pd.DataFrame({
            "UMAP-1": embedding[:, 0],
            "UMAP-2": embedding[:, 1],
            "UMAP-3": embedding[:, 2],
            color: groups,
        })
        fig = px.scatter_3d(df,
            x='UMAP-1', y='UMAP-2', z='UMAP-3',
            color=color, width=width, height=height,
            color_discrete_sequence=px.colors.qualitative.Dark24,
        )
    else:
        df = pd.DataFrame({
            "UMAP-1": embedding[:, 0],
            "UMAP-2": embedding[:, 1],
            color: groups,
        })
        fig = px.scatter(
            df, x="UMAP-1", y="UMAP-2", color=color, width=width, height=height,
            color_discrete_sequence=px.colors.qualitative.Dark24,
        )
    fig.update_traces(
        marker_size=marker_size,
        marker={"opacity": marker_opacity},
    )

    fig.update_layout(
        template="simple_white",
        legend= {'itemsizing': 'constant'},
    )
    return render_plot(fig, interactive, show, out_file)


from ._basic import import_data, make_tile_matrix, make_peak_matrix, filter_cells, select_features, make_gene_matrix
from ._knn import knn
from ._mnn_correct import mnc_correct
from ._scrublet import scrublet, call_doublets
from ._harmony import harmony


import numpy as np

def binarize_inplace(X):
    """Binarize sparse matrix in-place"""
    X.data = np.ones(X.indices.shape, dtype=np.float64)

def get_binarized_matrix(X):
    """Return a copy of binarize sparse matrix"""
    X_ = X.copy()
    binarize_inplace(X_)
    return X_

def get_igraph_from_adjacency(adj):
    """Get igraph graph from adjacency matrix."""
    import igraph as ig
    vcount = max(adj.shape)
    sources, targets = adj.nonzero()
    edgelist = list(zip(list(sources), list(targets)))
    weights = np.ravel(adj[(sources, targets)])
    gr = ig.Graph(n=vcount, edges=edgelist, edge_attrs={"weight": weights})
    return gr

def chunks(mat, chunk_size: int):
    """
    Return chunks of the input matrix
    """
    n = mat.shape[0]
    for i in range(0, n, chunk_size):
        j = max(i + chunk_size, n)
        yield mat[i:j, :]


"""
Use harmony to integrate cells from different experiments.
"""

import numpy as np
from typing import Optional, List, Union

from snapatac2._snapatac2 import AnnData, AnnDataSet

def harmony(
    adata: AnnData,
    batch: str,
    use_dims: Optional[Union[int, List[int]]] = None,
    use_rep: Optional[str] = None,
    inplace: bool = True,
    **kwargs,
) -> Optional[np.ndarray]:
    """
    Use harmonypy to integrate different experiments.

    Harmony is an algorithm for integrating single-cell
    data from multiple experiments. This function uses the python
    port of Harmony, ``harmonypy``, to integrate single-cell data
    stored in an AnnData object. This function should be run after performing
    dimension reduction.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    batch
        The name of the column in ``adata.obs`` that differentiates
        among experiments/batches.
    use_dims
        Use these dimensions in `use_rep`.
    use_rep
        The name of the field in ``adata.obsm`` where the lower dimensional
        representation is stored. Defaults to ``'X_spectral'``.
    inplace
        Whether to store the result in the anndata object.
    kwargs
        Any additional arguments will be passed to
        ``harmonypy.run_harmony()``.

    Returns
    -------
    if `inplace=True` it updates adata with the field
    ``adata.obsm[`use_rep`_harmony]``, containing principal components
    adjusted by Harmony such that different experiments are integrated.
    Otherwise, it returns the result as a numpy array.
    """
    try:
        import harmonypy
    except ImportError:
        raise ImportError("\nplease install harmonypy:\n\n\tpip install harmonypy")

    if use_rep is None: use_rep = "X_spectral"
    mat = adata.obsm[use_rep] if isinstance(adata, AnnData) or isinstance(adata, AnnDataSet) else adata
    if isinstance(use_dims, int): use_dims = range(use_dims) 
    mat = mat if use_dims is None else mat[:, use_dims]

    harmony_out = harmonypy.run_harmony(mat, adata.obs[...].to_pandas(), batch, **kwargs)
    if (isinstance(adata, AnnData) or isinstance(adata, AnnDataSet)) and inplace:
        adata.obsm[use_rep + "_harmony"] = harmony_out.Z_corr.T
    else:
        return harmony_out.Z_corr.T


import numpy as np
from sklearn.metrics.pairwise import cosine_similarity, rbf_kernel

from typing import Optional, Union
from scipy.sparse.linalg import svds

from snapatac2._snapatac2 import AnnData
from ._spectral import JaccardNormalizer, jaccard_similarity

def laplacian(
    data: AnnData,
    n_comps: Optional[int] = None,
    features: Optional[Union[str, np.ndarray]] = "selected",
    random_state: int = 0,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Convert chromatin accessibility profiles of cells into lower dimensional representations.

    Convert chromatin accessibility profiles of cells into lower dimensional representations
    using the spectrum of the normalized graph Laplacian defined by pairwise similarity
    between cells.

    Parameters
    ----------
    data
        AnnData object
    n_comps
        Number of dimensions to keep
    features
        Boolean index mask. True means that the feature is kept.
        False means the feature is removed.
    random_state
        Seed of the random state generator

    Returns
    -------
    if `inplace=True` it stores Spectral embedding of data in the field
    `adata.obsm["X_spectral"]`, otherwise it returns the result as a numpy array.
    """
    features = data.var[features] if isinstance(features, str) else features
    if n_comps is None:
        min_dim = min(data.n_vars, data.n_obs)
        if 50 >= min_dim:
            n_comps = min_dim - 1
        else:
            n_comps = 50
    (n_sample, _) = data.shape

    X = data.X[...]
    X.data = np.ones(X.indices.shape, dtype=np.float64)

    if features is not None: X = X[:, features]

    result = _laplacian(X, n_comps, distance="jaccard")
    if inplace:
        data.obsm['X_spectral'] = result
    else:
        return result

def _laplacian(mat, n_dim, distance="jaccard"):
    if distance == "jaccard":
        compute_similarity = jaccard_similarity
    elif distance == "cosine":
        compute_similarity = cosine_similarity
    else:
        compute_similarity = rbf_kernel

    dim = mat.shape[1]
    coverage = mat.sum(axis=1) / dim
    print("Compute similarity matrix")
    A = compute_similarity(mat)
    if distance == "jaccard":
        print("Normalization")
        normalizer = JaccardNormalizer(A, coverage)
        normalizer.normalize(A, coverage, coverage)
        np.fill_diagonal(A, 0)
        # Remove outlier
        normalizer.outlier = np.quantile(np.asarray(A), 0.999)
        np.clip(A, a_min=0, a_max=normalizer.outlier, out=A)
    else:
        np.fill_diagonal(A, 0)

    np.divide(A, A.sum(axis=1), out=A)

    print("Perform decomposition")
    u, s, _ = svds(A, k = n_dim, return_singular_vectors="u")
    ix = s.argsort()[::-1]
    s = s[ix]
    u = u[:, ix]
    return np.multiply(u, s)


from typing import Optional, Union, Type
import numpy as np
import polars

from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as _snapatac2

def kmeans(
    adata: Union[AnnData, AnnDataSet, np.ndarray],
    n_clusters: int,
    n_iterations: int = -1,
    random_state: int = 0,
    use_rep: Optional[str] = None,
    key_added: str = 'kmeans',
    inplace: bool = True,
) -> None:
    """
    Cluster cells into subgroups using the K-means algorithm.

    Parameters
    ----------
    adata
        The annotated data matrix.
    n_clusters
        Number of clusters to return.
    n_iterations
        How many iterations of the kmeans clustering algorithm to perform.
        Positive values above 2 define the total number of iterations to perform,
        -1 has the algorithm run until it reaches its optimal clustering.
    random_state
        Change the initialization of the optimization.
    use_rep
        Which data in `adata.obsm` to use for clustering. Default is "X_spectral".
    key_added
        `adata.obs` key under which to add the cluster labels.

    Returns
    -------
    adds fields to `adata`:
    `adata.obs[key_added]`
        Array of dim (number of samples) that stores the subgroup id
        (`'0'`, `'1'`, ...) for each cell.
    `adata.uns['kmeans']['params']`
        A dict with the values for the parameters `n_clusters`, `random_state`,
        and `n_iterations`.
    """
    if use_rep is None: use_rep = "X_spectral"

    if isinstance(adata, AnnData) or isinstance(adata, AnnDataSet):
        data = adata.obsm[use_rep]
    else:
        data = adata
    groups = _snapatac2.kmeans(n_clusters, data)
    groups = np.array(groups, dtype=np.compat.unicode)
    if inplace:
        adata.obs[key_added] = polars.Series(
            groups,
            dtype=polars.datatypes.Categorical,
        )
        # store information on the clustering parameters
        #adata.uns['kmeans'] = {}
        #adata.uns['kmeans']['params'] = dict(
        #    n_clusters=n_clusters,
        #    random_state=random_state,
        #    n_iterations=n_iterations,
        #)

    else:
        return groups


import scipy as sp
import numpy as np
from sklearn.metrics.pairwise import rbf_kernel
import gc
from typing import Optional, Union

from snapatac2._snapatac2 import AnnData, AnnDataSet, jm_regress, jaccard_similarity, cosine_similarity

def idf(data, features=None):
    n, m = data.shape
    count = np.zeros(m)
    for batch in data.X.chunked(2000):
        batch.data = np.ones(batch.indices.shape, dtype=np.float64)
        count += np.ravel(batch.sum(axis = 0))
    if features is not None:
        count = count[features]
    return np.log(n / (1 + count))

# FIXME: random state
def spectral(
    data: Union[AnnData, AnnDataSet],
    n_comps: int = 50,
    features: Optional[Union[str, np.ndarray]] = "selected",
    random_state: int = 0,
    sample_size: Optional[Union[int, float]] = None,
    chunk_size: int = 20000,
    distance_metric: str = "jaccard",
    feature_weights: Optional[Union[str, np.ndarray]] = "idf",
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Compute Laplacian Eigenmaps of chromatin accessibility profiles.

    Convert chromatin accessibility profiles of cells into lower dimensional representations
    using the spectrum of the normalized graph Laplacian defined by pairwise similarity
    between cells.

    Parameters
    ----------
    data
        AnnData or AnnDataSet object.
    n_comps
        Number of dimensions to keep.
    features
        Boolean index mask. True means that the feature is kept.
        False means the feature is removed.
    random_state
        Seed of the random state generator
    sample_size
        Sample size used in the Nystrom method. It could be either an integer
        indicating the number of cells to sample or a real value from 0 to 1
        indicating the fraction of cells to sample.
    chunk_size
        Chunk size used in the Nystrom method
    distance_metric
        distance metric: "jaccard", "cosine".
    feature_weights
        Whether to weight features using "IDF".
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    if `inplace=True` it stores Spectral embedding of data in the field
    `adata.obsm["X_spectral"]`,
    `adata.uns["spectral_eigenvalue"]`,
    otherwise it returns the result as a numpy array.
    """
    if not (isinstance(data, AnnData) or isinstance(data, AnnDataSet)):
        raise ValueError("input should be AnnData or AnnDataSet")

    if isinstance(features, str):
        if features in data.var:
            features = data.var[features]
        else:
            raise NameError("Please call `select_features` first or explicitly set `features = None`")

    if feature_weights is None:
        weights = None
    elif isinstance(feature_weights, np.ndarray):
        weights = feature_weights if features is None else feature_weights[features]
    elif feature_weights == "idf":
        weights = idf(data, features)
    else:
        raise NameError("Invalid feature_weights")

    np.random.seed(random_state)
    n_comps = min(data.n_vars - 1, data.n_obs - 1, n_comps)

    (n_sample, _) = data.shape
    if sample_size is None:
        sample_size = n_sample
    elif isinstance(sample_size, int):
        if sample_size <= 1:
            raise ValueError("when sample_size is an integer, it should be > 1")
        if sample_size > n_sample:
            sample_size = n_sample
    else:
        if sample_size <= 0.0 or sample_size > 1.0:
            raise ValueError("when sample_size is a float, it should be > 0 and <= 1")
        else:
            sample_size = int(sample_size * n_sample)

    if sample_size >= n_sample:
        if features is not None:
            X = data.X[:, features]
        else:
            X = data.X[...]
        model = Spectral(n_comps, distance_metric, weights)
        model.fit(X)
        result = model.transform()
    else:
        S = data.X.chunk(sample_size, replace=False)
        if features is not None: S = S[:, features]

        model = Spectral(n_comps, distance_metric, weights)
        model.fit(S)

        from tqdm import tqdm
        import math
        print("Perform Nystrom extension")
        chunks_iter = data.X.chunked(chunk_size)
        for batch in tqdm(chunks_iter, total = chunks_iter.n_chunks()):
            if distance_metric == "jaccard":
                batch.data = np.ones(batch.indices.shape, dtype=np.float64)
            if features is not None: batch = batch[:, features]
            model.extend(batch)
        result = model.transform()

    if inplace:
        data.uns['spectral_eigenvalue'] = result[0]
        data.obsm['X_spectral'] = result[1]
    else:
        return result


class Spectral:
    def __init__(self, n_dim: int = 30, distance: str = "jaccard", feature_weights = None):

        #self.dim = mat.get_shape()[1]
        self.n_dim = n_dim
        self.distance = distance
        if (self.distance == "jaccard"):
            self.compute_similarity = lambda x, y=None: jaccard_similarity(x, y, feature_weights)
        elif (self.distance == "cosine"):
            self.compute_similarity = lambda x, y=None: cosine_similarity(x, y, feature_weights)
        else:
            self.compute_similarity = rbf_kernel

    def fit(self, mat, verbose: int = 1):
        self.sample = mat
        self.dim = mat.shape[1]
        self.coverage = mat.sum(axis=1) / self.dim
        if verbose > 0: print("Compute similarity matrix")
        A = self.compute_similarity(mat)

        if (self.distance == "jaccard"):
            if verbose > 0: print("Normalization")
            self.normalizer = JaccardNormalizer(A, self.coverage)
            self.normalizer.normalize(A, self.coverage, self.coverage)
            np.fill_diagonal(A, 0)
            # Remove outlier
            self.normalizer.outlier = np.quantile(A, 0.999)
            np.clip(A, a_min=0, a_max=self.normalizer.outlier, out=A)
        else:
            np.fill_diagonal(A, 0)

        # M <- D^-1/2 * A * D^-1/2
        D = np.sqrt(A.sum(axis=1)).reshape((-1, 1))
        np.divide(A, D, out=A)
        np.divide(A, D.T, out=A)

        if verbose > 0: print("Perform decomposition")
        evals, evecs = sp.sparse.linalg.eigsh(A, self.n_dim + 1, which='LM')
        ix = evals.argsort()[::-1]
        self.evals = np.real(evals[ix])
        self.evecs = np.real(evecs[:, ix])

        B = np.divide(self.evecs, D)
        np.divide(B, self.evals.reshape((1, -1)), out=B)

        self.B = B
        self.Q = []

        return self

    def extend(self, data):
        A = self.compute_similarity(self.sample, data)
        if (self.distance == "jaccard"):
            self.normalizer.normalize(
                A, self.coverage, data.sum(axis=1) / self.dim,
                clip_min=0, clip_max=self.normalizer.outlier
            )
        self.Q.append(A.T @ self.B)

    def transform(self, orthogonalize = True):
        if len(self.Q) > 0:
            Q = np.concatenate(self.Q, axis=0)
            D_ = np.sqrt(np.multiply(Q, self.evals.reshape(1, -1)) @ Q.sum(axis=0).T)
            np.divide(Q, D_.reshape((-1, 1)), out=Q)

            if orthogonalize:
                # orthogonalization
                sigma, V = np.linalg.eig(Q.T @ Q)
                sigma = np.sqrt(sigma)
                B = np.multiply(V.T, self.evals.reshape((1,-1))) @ V
                np.multiply(B, sigma.reshape((-1, 1)), out=B)
                np.multiply(B, sigma.reshape((1, -1)), out=B)
                evals_new, evecs_new = np.linalg.eig(B)

                # reorder
                ix = evals_new.argsort()[::-1]
                self.evals = evals_new[ix]
                evecs_new = evecs_new[:, ix]

                np.divide(evecs_new, sigma.reshape((-1, 1)), out=evecs_new)
                self.evecs = Q @ V @ evecs_new
            else:
                self.evecs = Q
        return (self.evals[1:], self.evecs[:, 1:])

class JaccardNormalizer:
    def __init__(self, jm, c):
        (slope, intersect) = jm_regress(jm, c)
        self.slope = slope
        self.intersect = intersect
        self.outlier = None

    def normalize(self, jm, c1, c2, clip_min=None, clip_max=None):
        # jm / (self.slope / (1 / c1 + 1 / c2.T - 1) + self.intersect)
        temp = 1 / c1 + 1 / c2.T
        temp -= 1
        np.reciprocal(temp, out=temp)
        np.multiply(temp, self.slope, out=temp)
        temp += self.intersect
        jm /= temp
        if clip_min is not None or clip_max is not None:
            np.clip(jm, a_min=clip_min, a_max=clip_max, out=jm)
        gc.collect()


"""
def nystrom_full(mat, sample_size, n_dim):
    n, m = mat.shape
    sample_indices = np.random.choice(n, size=sample_size, replace=False)
    mask = np.ones(n, bool)
    mask[sample_indices] = False
    sample = mat[sample_indices, :]
    other_data = mat[mask, :]

    # Compute affinity matrix
    coverage = sample.sum(axis=1) / m
    A = jaccard_similarity(sample)
    normalizer = JaccardNormalizer(A, coverage)
    normalizer.normalize(A, coverage, coverage)
    np.fill_diagonal(A, 0)
    normalizer.outlier = np.quantile(np.asarray(A), 0.999)
    np.clip(A, a_min=0, a_max=normalizer.outlier, out=A)

    # Compute distance matrix B
    B = jaccard_similarity(sample, other_data)
    normalizer.normalize(B, coverage, other_data.sum(axis=1) / m,
        clip_min=0, clip_max=normalizer.outlier)

    # Compute degree
    a_r = A.sum(axis=1)
    b_r = B.sum(axis=1)
    b_c = B.sum(axis=0).reshape((-1, 1))
    d1 = np.sqrt(a_r + b_r)
    d2 = np.sqrt(np.clip(b_c + B.T @ np.linalg.pinv(A) @ b_r, a_min=1e-10, a_max=None))

    # normalization
    np.divide(A, d1, out=A)
    np.divide(A, d1.T, out=A)
    np.divide(B, d1.reshape((-1, 1)), out=B)
    np.divide(B, d2.reshape((1, -1)), out=B)

    # compute eigenvector
    evals, U = sp.sparse.linalg.eigsh(A, n_dim + 1, which='LM')
    U_ = np.divide(B.T @ U, evals.reshape((1, -1)))
    result = np.empty((n, n_dim))
    result[sample_indices] = U[:, 1:]
    result[mask] = U_[:, 1:]
    return result
"""


#from ._io import read_10x_mtx
from . import preprocessing as pp
from . import tools as tl
from . import genome
from . import plotting as pl
from . import export as ex

from snapatac2._snapatac2 import AnnData, AnnDataSet, read, read_mtx, create_dataset, read_dataset

import sys
sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['pp', 'tl', 'pl', 'ex']})
del sys


from ._qc import tsse, scrublet
from ._scatter import umap

from typing import Optional
from snapatac2._snapatac2 import AnnData
from ._utils import render_plot

def detect(x, saturation=0.01):
    accum_gap = 0
    for i in range(1, len(x)):
        gap = x[i-1] - x[i]
        accum_gap = accum_gap + gap
        if gap < saturation * accum_gap:
            return i
    return None

def spectral_eigenvalues(
    adata: AnnData,
    n_components: Optional[int] = None,
    show: bool = True,
    interactive: bool = True,
    out_file: Optional[str] = None,
    **kwargs,
) -> None:
    import plotly.express as px
    import pandas as pd

    data = adata.uns["spectral_eigenvalue"]
    n = data.shape[0] if n_components is None else n_components
    data = data[:n]

    df = pd.DataFrame({"Component": map(str, range(1, n+1)), "Eigenvalue": data})
    fig = px.scatter(df, x = "Component", y = "Eigenvalue", template="plotly_white", **kwargs)
    fig.add_vline(x=detect(data))

    return render_plot(fig, interactive, show, out_file)


from typing import Optional, Union
import scipy.sparse as ss
import numpy as np
import polars

from snapatac2._snapatac2 import AnnData, AnnDataSet
from .. import _utils

def leiden(
    adata: Union[AnnData, AnnDataSet, ss.spmatrix],
    resolution: float = 1,
    objective_function: str = "modularity",
    min_cluster_size: int = 5,
    n_iterations: int = -1,
    random_state: int = 0,
    key_added: str = 'leiden',
    use_leidenalg: bool = False,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Cluster cells into subgroups [Traag18]_.

    Cluster cells using the Leiden algorithm [Traag18]_,
    an improved version of the Louvain algorithm [Blondel08]_.
    It has been proposed for single-cell analysis by [Levine15]_.
    This requires having ran :func:`~snapatac2.pp.knn`.

    Parameters
    ----------
    adata
        The annotated data matrix or sparse adjacency matrix of the graph,
        defaults to neighbors connectivities.
    resolution
        A parameter value controlling the coarseness of the clustering.
        Higher values lead to more clusters.
        Set to `None` if overriding `partition_type`
        to one that doesn't accept a `resolution_parameter`.
    objective_function
        whether to use the Constant Potts Model (CPM) or modularity.
        Must be either "CPM", "modularity" or "RBConfiguration".
    min_cluster_size
        The minimum size of clusters.
    n_iterations
        How many iterations of the Leiden clustering algorithm to perform.
        Positive values above 2 define the total number of iterations to perform,
        -1 has the algorithm run until it reaches its optimal clustering.
    random_state
        Change the initialization of the optimization.
    key_added
        `adata.obs` key under which to add the cluster labels.
    use_leidenalg
        If `True`, `leidenalg` package is used. Otherwise, `python-igraph` is used.
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    adds fields to `adata`:
    `adata.obs[key_added]`
        Array of dim (number of samples) that stores the subgroup id
        (`'0'`, `'1'`, ...) for each cell.
    `adata.uns['leiden']['params']`
        A dict with the values for the parameters `resolution`, `random_state`,
        and `n_iterations`.
    """
    from collections import Counter

    if isinstance(adata, AnnData) or isinstance(adata, AnnDataSet):
        adjacency = adata.obsp["distances"]
    else:
        adjacency = adata
    
    gr = _utils.get_igraph_from_adjacency(adjacency)

    if use_leidenalg or objective_function == "RBConfiguration":
        import leidenalg
        from leidenalg.VertexPartition import MutableVertexPartition

        if objective_function == "modularity":
            partition_type = leidenalg.ModularityVertexPartition
        elif objective_function == "CPM":
            partition_type = leidenalg.CPMVertexPartition
        elif objective_function == "RBConfiguration":
            partition_type = leidenalg.RBConfigurationVertexPartition
        else:
            raise ValueError(
                'objective function is not supported: ' + partition_type
            )

        partition = leidenalg.find_partition(
            gr, partition_type, n_iterations=n_iterations,
            seed=random_state, resolution_parameter=resolution, weights=None
        )
    else:
        from igraph import set_random_number_generator
        import random
        random.seed(random_state)
        set_random_number_generator(random)
        partition = gr.community_leiden(
            objective_function=objective_function,
            weights=None,
            resolution_parameter=resolution,
            beta=0.01,
            initial_membership=None,
            n_iterations=n_iterations,
        )

    groups = partition.membership

    new_cl_id = dict([(cl, i) if count >= min_cluster_size else (cl, -1) for (i, (cl, count)) in enumerate(Counter(groups).most_common())])
    for i in range(len(groups)): groups[i] = new_cl_id[groups[i]]

    groups = np.array(groups, dtype=np.compat.unicode)
    if inplace:
        adata.obs[key_added] = polars.Series(
            groups,
            dtype=polars.datatypes.Categorical,
        )
        # store information on the clustering parameters
        #adata.uns['leiden'] = {}
        #adata.uns['leiden']['params'] = dict(
        #    resolution=resolution,
        #    random_state=random_state,
        #    n_iterations=n_iterations,
        #)
    else:
        return groups


from typing import Optional, Union, Type
import pandas as pd
from typing import Optional
import numpy as np

from snapatac2._snapatac2 import AnnData

def dbscan(
    adata: AnnData,
    eps: float = 0.5,
    min_samples: int = 5,
    leaf_size: int = 30,
    n_jobs: Optional[int] = None,
    use_rep: Optional[str] = None,
    key_added: str = 'dbscan',
) -> None:
    """
    Cluster cells into subgroups using the DBSCAN algorithm.

    Parameters
    ----------
    adata
        The annotated data matrix.
    eps
        The maximum distance between two samples for one to be considered as
        in the neighborhood of the other. This is not a maximum bound on the
        distances of points within a cluster. This is the most important
        DBSCAN parameter to choose appropriately for your data set and distance function.
    min_samples
        The number of samples (or total weight) in a neighborhood for a point
        to be considered as a core point. This includes the point itself.
    leaf_size
        Leaf size passed to BallTree or cKDTree. This can affect the speed of
        the construction and query, as well as the memory required to store the
        tree. The optimal value depends on the nature of the problem.
    n_jobs
        The number of parallel jobs to run. None means 1 unless in a
        joblib.parallel_backend context. -1 means using all processors.
    use_rep
        Which data in `adata.obsm` to use for clustering. Default is "X_spectral".
    key_added
        `adata.obs` key under which to add the cluster labels.

    Returns
    -------
    adds fields to `adata`:
    `adata.obs[key_added]`
        Array of dim (number of samples) that stores the subgroup id
        (`'0'`, `'1'`, ...) for each cell.
    """
    from sklearn.cluster import DBSCAN
    if use_rep is None: use_rep = "X_spectral"
    data = adata.obsm[use_rep][...]

    clustering = DBSCAN(
        eps=eps,
        min_samples=min_samples,
        metric='euclidean',
        leaf_size=leaf_size,
        n_jobs=n_jobs).fit(data)
    groups = clustering.labels_
    adata.obs[key_added] = pd.Categorical(
        values=groups.astype('U'),
        categories=sorted(map(str, np.unique(groups))),
    )


from typing import Optional, Union, Type
import pandas as pd
import numpy as np

from snapatac2._snapatac2 import AnnData

def hdbscan(
    adata: AnnData,
    min_cluster_size: int = 5,
    min_samples: Optional[int] = None,
    cluster_selection_epsilon: float = 0.0,
    alpha: float = 1.0,
    cluster_selection_method: str = "eom",
    random_state: int = 0,
    use_rep: Optional[str] = None,
    key_added: str = 'hdbscan',
    **kwargs,
) -> None:
    """
    Cluster cells into subgroups using the HDBSCAN algorithm.

    Parameters
    ----------
    adata
        The annotated data matrix.
    min_cluster_size
        The minimum size of clusters; single linkage splits that contain
        fewer points than this will be considered points "falling out" of
        a cluster rather than a cluster splitting into two new clusters.
    min_samples
        The number of samples in a neighbourhood for a point to be considered a core point.
    cluster_selection_epsilon
        A distance threshold. Clusters below this value will be merged.
    alpha
        A distance scaling parameter as used in robust single linkage.
    cluster_selection_method
        The method used to select clusters from the condensed tree.
        The standard approach for HDBSCAN* is to use an Excess of Mass
        algorithm to find the most persistent clusters.
        Alternatively you can instead select the clusters at the leaves of
        the tree - this provides the most fine grained and homogeneous clusters.
        Options are: "eom" or "leaf".
    random_state
        Change the initialization of the optimization.
    use_rep
        Which data in `adata.obsm` to use for clustering. Default is "X_spectral".
    key_added
        `adata.obs` key under which to add the cluster labels.

    Returns
    -------
    adds fields to `adata`:
    `adata.obs[key_added]`
        Array of dim (number of samples) that stores the subgroup id
        (`'0'`, `'1'`, ...) for each cell.
    """
    import hdbscan
    if use_rep is None: use_rep = "X_spectral"
    data = adata.obsm[use_rep][...]
    clusterer = hdbscan.HDBSCAN(
        min_cluster_size = min_cluster_size,
        min_samples = min_samples,
        cluster_selection_epsilon = cluster_selection_epsilon,
        alpha = alpha,
        cluster_selection_method = cluster_selection_method,
        **kwargs
    )
    clusterer.fit(data)
    groups = clusterer.labels_
    adata.obs[key_added] = pd.Categorical(
        values=groups.astype('U'),
        categories=sorted(map(str, np.unique(groups))),
    )


from pathlib import Path
import numpy as np
import math
from typing import Optional, Union, Mapping, Set

from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as internal

def import_data(
    fragment_file: Path,
    gff_file: Path,
    chrom_size: Mapping[str, int],
    file: Path,
    min_num_fragments: int = 200,
    min_tsse: float = 1,
    sorted_by_barcode: bool = True,
    whitelist: Optional[Union[Path, Set[str]]] = None,
    chunk_size: int = 2000,
    n_jobs: int = 4,
) -> AnnData:
    """
    Import dataset and compute QC metrics.

    This function will store fragments as base-resolution TN5 insertions in the
    resulting h5ad file (in `.obsm['insertion']`), along with the chromosome
    sizes (in `.uns['reference_sequences']`). Various QC metrics, including TSSe,
    number of unique fragments, duplication rate, fraction of mitochondrial DNA
    reads, will be computed.

    Parameters
    ----------
    fragment_file
        File name of the fragment file
    gff_file
        File name of the gene annotation file in GFF format
    chrom_size
        A dictionary containing chromosome sizes, for example,
        `{"chr1": 2393, "chr2": 2344, ...}`
    file
        File name of the output h5ad file used to store the result
    min_num_fragments
        Number of unique fragments threshold used to filter cells
    min_tsse
        TSS enrichment threshold used to filter cells
    sorted_by_barcode
        Whether the fragment file has been sorted by cell barcodes. Pre-sort the
        fragment file will speed up the processing and require far less memory.
    whitelist
        File name or a set of strings. If it is a file name, each line
        must contain a valid barcode. 
    chunk_size
        chunk size
    n_jobs
        number of CPUs to use

    Returns
    -------
    An annotated data matrix of shape `n_obs` x `n_vars`.
    Rows correspond to cells and columns to regions.
    """
    if isinstance(whitelist, str) or isinstance(whitelist, Path):
        with open(whitelist, "r") as fl:
            whitelist = set([line.strip() for line in fl])
    return internal.import_fragments(
        str(file), str(fragment_file), str(gff_file), chrom_size,
        min_num_fragments, min_tsse, sorted_by_barcode, whitelist, chunk_size, n_jobs
    )

def make_tile_matrix(
    adata: AnnData,
    bin_size: int = 500,
    chunk_size: int = 500,
    n_jobs: int = 4
):
    """
    Generate cell by bin count matrix.

    This function will generate and add a cell by bin count matrix to the AnnData
    object in place.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    bin_size
        The size of consecutive genomic regions used to record the counts.
    n_jobs
        number of CPUs to use.
    """
    internal.mk_tile_matrix(adata, bin_size, chunk_size, n_jobs)

def make_peak_matrix(
    adata: Union[AnnData, AnnDataSet],
    file: Path,
    use_rep: str = "peaks",
    peak_file: Optional[Path] = None,
) -> AnnData:
    """
    Generate cell by peak count matrix.

    This function will create a new .h5ad file to store the cell by peak count matrix.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    file
        File name of the h5ad file used to store the result.
    use_rep
        This is used to read peak information from `.uns[use_rep]`.
    peak_file
        Bed file containing the peaks. If provided, peak information will be read
        from this file.
    """
    peak_file = peak_file if peak_file is None else str(peak_file)
    anndata = internal.mk_peak_matrix(adata, use_rep, peak_file, str(file))
    anndata.obs = adata.obs[...]
    return anndata

def make_gene_matrix(
    adata: Union[AnnData, AnnDataSet],
    gff_file: Path,
    file: Path,
    chunk_size: int = 500,
    use_x: bool = False,
) -> AnnData:
    """
    Generate cell by gene activity matrix.

    Generate cell by gene activity matrix by counting the TN5 insertions in gene
    body regions. The result will be stored in a new file and a new AnnData object
    will be created.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    gff_file
        File name of the gene annotation file in GFF format.
    file
        File name of the h5ad file used to store the result.
    use_x
        If True, use the matrix stored in `.X` to compute the gene activity.
        Otherwise the base-resolution TN5 insertions are used.

    Returns
    -------
    A new AnnData object, where rows correspond to cells and columns to genes.
    """
    anndata = internal.mk_gene_matrix(adata, str(gff_file), str(file), chunk_size, use_x)
    anndata.obs = adata.obs[...]
    return anndata

def filter_cells(
    data: AnnData,
    min_counts: Optional[int] = 1000,
    min_tsse: Optional[float] = 5.0,
    max_counts: Optional[int] = None,
    max_tsse: Optional[float] = None,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Filter cell outliers based on counts and numbers of genes expressed.
    For instance, only keep cells with at least `min_counts` counts or
    `min_tsse` TSS enrichment scores. This is to filter measurement outliers,
    i.e. "unreliable" observations.

    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    min_counts
        Minimum number of counts required for a cell to pass filtering.
    min_tsse
        Minimum TSS enrichemnt score required for a cell to pass filtering.
    max_counts
        Maximum number of counts required for a cell to pass filtering.
    max_tsse
        Maximum TSS enrichment score expressed required for a cell to pass filtering.
    inplace
        Perform computation inplace or return result.

    Returns
    -------
    If `inplace = True`, directly subsets the data matrix. Otherwise return 
    a boolean index mask that does filtering, where `True` means that the
    cell is kept, `False` means the cell is removed.
    """
    selected_cells = True
    if min_counts: selected_cells &= data.obs["n_fragment"] >= min_counts
    if max_counts: selected_cells &= data.obs["n_fragment"] <= max_counts
    if min_tsse: selected_cells &= data.obs["tsse"] >= min_tsse
    if max_tsse: selected_cells &= data.obs["tsse"] <= max_tsse

    if inplace:
        data.subset(selected_cells)
    else:
        return selected_cells
 
def select_features(
    adata: Union[AnnData, AnnDataSet],
    most_variable: Optional[Union[int, float]] = 1000000,
    whitelist: Optional[str] = None,
    blacklist: Optional[str] = None,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Perform feature selection.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    most_variable
        If None, do not perform feature selection using most variable features
    whitelist
        A user provided bed file containing genome-wide whitelist regions.
        Features that are overlapped with these regions will be retained.
    blacklist 
        A user provided bed file containing genome-wide blacklist regions.
        Features that are overlapped with these regions will be removed.
    inplace
        Perform computation inplace or return result.
    
    Returns
    -------
    If `inplace = False`, return a boolean index mask that does filtering,
    where `True` means that the feature is kept, `False` means the feature is removed.
    Otherwise, store this index mask directly to `.var['selected']`.

    Notes
    -----
    This function doesn't perform actual subsetting. The feature mask is used by
    various functions to generate submatrices on the fly.
    """
    count = np.zeros(adata.shape[1])
    for batch in adata.X.chunked(2000):
        count += np.ravel(batch.sum(axis = 0))

    selected_features = count != 0

    if whitelist is not None:
        selected_features &= internal.intersect_bed(list(adata.var_names), whitelist)
    if blacklist is not None:
        selected_features &= np.logical_not(internal.intersect_bed(list(adata.var_names), blacklist))

    if most_variable is not None and len(count[selected_features]) > most_variable:
        mean = count[selected_features].mean()
        std = math.sqrt(count[selected_features].var())
        zscores = np.absolute((count - mean) / std)
        cutoff = np.sort(zscores)[most_variable - 1]
        selected_features &= zscores <= cutoff

    if inplace:
        adata.var["selected"] = selected_features
    else:
        return selected_features


def render_plot(fig, interactive: bool, show: bool, out_file: str):
    # save figure to file
    if out_file is not None:
        if out_file.endswith(".html"):
            fig.write_html(out_file, include_plotlyjs="cdn")
        else:
            fig.write_image(out_file)

    # show figure
    if show:
        if interactive:
            fig.show()
        else:
            from IPython.display import Image
            return Image(fig.to_image(format="png"))

    # return plot object
    if not show and not out_file: return fig


from typing import Optional, Union, List
import numpy as np
from scipy.sparse import csr_matrix

from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as internal

def knn(
    adata: Union[AnnData, AnnDataSet, np.ndarray],
    n_neighbors: int = 50,
    use_dims: Optional[Union[int, List[int]]] = None,
    use_rep: Optional[str] = None,
    use_approximate_search: bool = True,
    n_jobs: int = -1,
    inplace: bool = True,
) -> Optional[csr_matrix]:
    """
    Compute a neighborhood graph of observations.

    Parameters
    ----------
    adata
        Annotated data matrix or numpy array.
    n_neighbors
        The number of nearest neighbors to be searched.
    use_dims
        The dimensions used for computation.
    use_rep
        The key for the matrix
    use_approximate_search
        Whether to use approximate nearest neighbor search
    n_jobs
        number of CPUs to use
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    if `inplace`, store KNN in `.obsp['distances']`. Otherwise, return a sparse
    matrix.
    """
    if isinstance(adata, AnnData) or isinstance(adata, AnnDataSet):
        if use_rep is None: use_rep = "X_spectral"
        data = adata.obsm[use_rep]
    else:
        data = adata

    if use_dims is not None:
        if isinstance(use_dims, int):
            data = data[:, :use_dims]
        else:
            data = data[:, use_dims]

    n = data.shape[0]
    if use_approximate_search:
        (d, indices, indptr) = internal.approximate_nearest_neighbors(data.astype(np.float32), n_neighbors)
        adj = csr_matrix((d, indices, indptr), shape=(n, n))
    else:
        from sklearn.neighbors import kneighbors_graph
        adj = kneighbors_graph(data, n_neighbors, mode='distance', n_jobs=n_jobs)
    
    if inplace:
        adata.obsp['distances'] = adj
    else:
        return adj


import numpy as np
from scipy.stats import chi2
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import log_loss

def diff_analysis(mat, z, peaks=None, covariate=None):
    """
    Parameters
    ----------
    mat
        cell by peak matrix.
    z
        variables to test
    peaks
        peak indices
    covariate 
        additional variables to regress out.
    """

    if len(z.shape) == 1:
        z = z.reshape((-1, 1))
    
    depth_per_cell = np.log(np.sum(mat, axis=1))
    if covariate is None:
        X = depth_per_cell
    else:
        X = np.concatenate((depth_per_cell, covariate), axis=1)

    if peaks is None:
        mat = mat.tocsc()
    else:
        mat = mat.tocsc()[:, peaks]

    return _diff_test_helper(np.asarray(X), np.asarray(z), mat)


def _diff_test_helper(X, z, Y):
    """
    Parameters
    ----------
    X
        (n_sample, n_feature),
    z
        (n_sample, 1), additional variables
    Y
        (n_sample, k), labels
    """
    from tqdm import tqdm
 
    X0 = X
    X1 = np.concatenate((X, z), axis=1)

    _, n = Y.shape
    Y.data = np.ones(Y.data.shape)

    result = []
    for i in tqdm(range(n)):
        result.append(
            likelihood_ratio_test(X0, X1, np.asarray(np.ravel(Y[:, i].todense())))
        )
    return result

def likelihood_ratio_test(X0, X1, y) -> float:
    """
    Comparing null model with alternative model using the likehood ratio test.

    Parameters
    ----------
    X0
        (n_sample, n_feature), variables used in null model
    X1
        (n_sample, n_feature2), variables used in alternative model
    Y
        (n_sample, ), labels
    """
    model = LogisticRegression(penalty="none", random_state=0, n_jobs=1,
        solver="lbfgs", multi_class='ovr', warm_start=False,
        max_iter = 1000,
        ).fit(X0, y)
    reduced = -log_loss(y, model.predict_proba(X0), normalize=False)

    model = LogisticRegression(penalty="none", random_state=0, n_jobs=1,
        solver="lbfgs", multi_class='ovr', warm_start=False,
        max_iter = 1000,
        ).fit(X1, y)
    full = -log_loss(y, model.predict_proba(X1), normalize=False)
    chi = -2 * (reduced - full)
    return chi2.sf(chi, X1.shape[1] - X0.shape[1])


import numpy as np
from typing import Optional, Union, List

from snapatac2._snapatac2 import AnnData, AnnDataSet

def umap(
    adata: Union[AnnData, AnnDataSet, np.ndarray],
    n_comps: int = 2,
    use_dims: Optional[Union[int, List[int]]] = None,
    use_rep: Optional[str] = None,
    key_added: str = 'umap',
    random_state: int = 0,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    Parameters
    ----------
    adata
        The annotated data matrix.
    n_comps
        The number of dimensions of the embedding.
    use_dims
        Use these dimensions in `use_rep`.
    use_rep
        Use the indicated representation in `.obsm`.
    key_added
        `adata.obs` key under which to add the cluster labels.
    random_state
        Random seed.
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    None
    """
    from umap import UMAP

    if isinstance(adata, AnnData) or isinstance(adata, AnnDataSet):
        if use_rep is None: use_rep = "X_spectral"
        data = adata.obsm[use_rep]
    else:
        data = adata

    if use_dims is not None:
        if isinstance(use_dims, int):
            data = data[:, :use_dims]
        else:
            data = data[:, use_dims]

    umap = UMAP(
        random_state=random_state, n_components=n_comps
        ).fit_transform(data)
    if inplace:
        adata.obsm["X_" + key_added] = umap
    else:
        return umap


from ._leiden import leiden
from ._kmeans import kmeans
from ._dbscan import dbscan
from ._hdbscan import hdbscan
from ._spectral import spectral
from ._laplacian import laplacian
from ._umap import umap
from ._counts import aggregate_X
from ._call_peaks import call_peaks
from ._diff import diff_analysis


from typing import Optional
from snapatac2._snapatac2 import AnnData
import numpy as np
from ._utils import render_plot

def tsse(
    adata: AnnData,
    show_cells: bool = False,
    show: bool = True,
    interactive: bool = True,
    out_file: Optional[str] = None,
    min_fragment: int = 500,
):
    """
    Plot the TSS enrichment vs. number of fragments density figure.

    Parameters
    ----------
    adata
        Annotated data matrix.
    show_cells
        Whether to show individual cells as dots on the plot
    show
        Show the figure
    interactive
        Whether to make interactive plot
    out_file
        Path of the output file for saving the output image, end with '.svg' or '.pdf' or '.png'
    min_fragment
        The cells' unique fragments lower than it should be removed

    Returns
    -------
    
    """
    import plotly.graph_objects as go

    selected_cells = adata.obs["n_fragment"] >= min_fragment
    x = adata.obs["n_fragment"][selected_cells]
    y = adata.obs["tsse"][selected_cells]

    log_x = np.log10(x)
    log_x_min, log_x_max = log_x.min(), log_x.max()
    x_range = log_x_max - log_x_min
    bin_x = np.linspace(log_x_min - 0.1 * x_range, log_x_max + 0.2 * x_range, 20)

    y_min, y_max = y.min(), y.max()
    y_range = y_max - y_min
    bin_y = np.linspace(y_min - 0.1 * y_range, y_max + 0.2 * y_range, 15)

    (z, rx, ry) = np.histogram2d(log_x,y, bins=(bin_x, bin_y))
    fig = go.Figure(data =
        go.Contour(
            z = z.T,
            x = 10**rx,
            y = ry,
            colorscale = 'Blues',
        )
    )

    if show_cells:
        fig.add_trace(go.Scatter(
                x = x,
                y = y,
                xaxis = 'x',
                yaxis = 'y',
                mode = 'markers',
                marker = dict(color = 'rgba(0,0,0,0.3)', size = 3)
        ))

    fig.update_xaxes(type="log")
    fig.update_layout(
        template="simple_white",
        xaxis_title="Number of unique fragments",
        yaxis_title="TSS enrichment score",
    )

    return render_plot(fig, interactive, show, out_file)

def scrublet(
    adata: AnnData,
    show: bool = True,
    interactive: bool = True,
    out_file: Optional[str] = None,
):
    """
    Plot doublets

    Parameters
    ----------
    adata
        Annotated data matrix.
    show
        Show the figure
    interactive
        Whether to make interactive plot
    out_file
        Path of the output file for saving the output image, end with '.svg' or '.pdf' or '.png'

    Returns
    -------
    
    """
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go

    doublet_scores = adata.obs["doublet_score"]
    sim_scores = adata.uns["scrublet_sim_doublet_score"]

    thres = adata.uns["scrublet_threshold"] if "scrublet_threshold" in adata.uns else None

    if thres is None:
        title1 = "Observed cells"
        title2 = "Simulated doublets"
    else:
        p1 = (doublet_scores >= thres).sum() / doublet_scores.size
        p2 = (sim_scores >= thres).sum() / sim_scores.size
        title1 = "Observed cells ({:.2%} doublets)".format(p1)
        title2 = "Simulated doublets ({:.2%} doublets)".format(p2)

    fig = make_subplots(rows=1, cols=2, subplot_titles=[title1, title2])

    fig.add_trace(go.Histogram(x=doublet_scores),row=1, col=1)
    if thres is not None:
        fig.add_vline(x=thres, line_width=3, line_dash="dash", line_color="green")
        fig.add_vrect(x0=thres, x1 = doublet_scores.max(), line_width=0, fillcolor="red", opacity=0.2)

    fig.add_trace(go.Histogram(x=sim_scores), row=1, col=2)
    if thres is not None:
        fig.add_vline(x=thres, line_width=3, line_dash="dash", line_color="green")
        fig.add_vrect(x0=thres, x1 = sim_scores.max(), line_width=0, fillcolor="red", opacity=0.2)

    fig.update(layout_showlegend=False)
    return render_plot(fig, interactive, show, out_file)


from setuptools import setup
from setuptools_rust import Binding, RustExtension

from pathlib import Path
root_directory = Path(__file__).parent
long_description = (root_directory / "README.md").read_text()

setup(
    name="snapatac2",
    description='SnapATAC: Single Nucleus Analysis Pipeline for ATAC-seq',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://kzhang.org/SnapATAC2/', 
    author='Kai Zhang',
    author_email='kai@kzhang.org',
    license='MIT',
    version="2.0.0",
    rust_extensions=[
        RustExtension("snapatac2._snapatac2", binding=Binding.PyO3),
    ],
    packages=[
        "snapatac2",
        "snapatac2.preprocessing",
        "snapatac2.tools",
        "snapatac2.plotting",
        "snapatac2.export",
    ],
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[
        "numpy>=1.16.0",
        "scipy>=1.4",
        "scikit-learn>=0.22",
        "tqdm>=4.62",
        "umap-learn>=0.3.10",
        "natsort",
        "pandas",
        "python-igraph",
        "polars",
        "pyarrow",
        "plotly>=5.6.0"
    ],
)



import numpy as np
from sklearn.neighbors import KDTree
import itertools
from scipy.special import logsumexp
from sklearn.cluster import KMeans
from typing import Optional, Union, List

from snapatac2._snapatac2 import AnnData, AnnDataSet

def mnc_correct(
    data: AnnData,
    batch: str,
    n_neighbors: int = 5,
    n_clusters: int = 40,
    use_dims: Optional[Union[int, List[int]]] = None,
    use_rep: Optional[str] = None,
    n_iter: int = 1,
    inplace: bool = True,
) -> Optional[np.ndarray]:
    """
    A modified MNN-Correct algorithm based on cluster centroid.

    Parameters
    ----------
    data
        Matrice or AnnData object. Matrices should be shaped like n_obs x n_vars.
    batch
    n_neighbors
        Number of mutual nearest neighbors.
    n_clusters
        Number of clusters
    use_dims
    use_rep
    n_iter
    inplace

    Returns
    -------
    if `inplace=True` it updates adata with the field
    ``adata.obsm[`use_rep`_mnn]``, containing adjusted principal components.
    Otherwise, it returns the result as a numpy array.
    """
    if use_rep is None: use_rep = "X_spectral"
    mat = data.obsm[use_rep] if isinstance(data, AnnData) or isinstance(data, AnnDataSet) else data
    if isinstance(use_dims, int): use_dims = range(use_dims) 
    mat = mat if use_dims is None else mat[:, use_dims]
    mat = np.asarray(mat)

    labels = data.obs[batch]
    label_uniq = list(set(labels))

    for _ in range(n_iter):
        batch_idx = []
        data_by_batch = []
        for label in label_uniq:
            idx = [i for i, x in enumerate(labels) if x == label]
            batch_idx.append(idx)
            data_by_batch.append(mat[idx,:])
        mat_ = mnc_correct_multi(data_by_batch, n_neighbors, n_clusters)
        new = np.array(mat, copy=True)
        idx = list(itertools.chain.from_iterable(batch_idx))
        for i in range(len(labels)):
            new[idx[i]] = mat_[i,:]
        mat = new

    if (isinstance(data, AnnData) or isinstance(data, AnnDataSet)) and inplace:
        data.obsm[use_rep + "_mnn"] = mat
    else:
        return mat

def mnc_correct_multi(datas, n_neighbors, n_clusters):
    data0 = datas[0]
    for i in range(1, len(datas)):
        data1 = datas[i]
        pdata0, pdata1 = mnc_correct_pair(data0, data1, n_clusters, n_neighbors)
        ratio = data0.shape[0] / (data0.shape[0] + data1.shape[0])
        data0_ = pdata0.project(data0, weight = 1 - ratio)
        data1_ = pdata1.project(data1, weight = ratio)
        data0 = np.concatenate((data0_, data1_), axis=0)
    return data0

def mnc_correct_pair(X, Y, n_clusters, n_neighbors, random_state=0):
    n_X = X.shape[0]
    n_Y = Y.shape[0]
    c_X = KMeans(n_clusters=min(n_clusters, n_X), random_state=random_state).fit(X).cluster_centers_
    c_Y = KMeans(n_clusters=min(n_clusters, n_Y), random_state=random_state).fit(Y).cluster_centers_

    tree_X = KDTree(c_X)
    tree_Y = KDTree(c_Y)

    # X by Y matrix
    m_X = tree_Y.query(c_X, k=min(n_neighbors, n_Y), return_distance=False)

    # Y by X matrix
    m_Y_ = tree_X.query(c_Y, k=min(n_neighbors, n_X), return_distance=False)
    m_Y = []
    for i in range(m_Y_.shape[0]):
        m_Y.append(set(m_Y_[i,:]))

    i_X = []
    i_Y = []
    for i in range(m_X.shape[0]):
        for j in m_X[i]:
            if i in m_Y[j]:
                i_X.append(i)
                i_Y.append(j)
    a = c_X[i_X,:]
    b = c_Y[i_Y,:]
    return (Projector(a, b), Projector(b, a))

class Projector(object):
    def __init__(self, X, Y):
        self.reference = X
        self.vector = Y - X

    def project(self, X, weight=0.5):
        def project(x):
            P = self.reference
            U = self.vector
            d = np.sqrt(np.sum((P - x)**2, axis=1))
            w = _normalize(-(d/0.005))
            #w = 1/d
            return (x + weight * np.average(U, axis=0, weights=w))
        return np.apply_along_axis(project, 1, X)

# exp transform the weights and then normalize them to sum to 1.
def _normalize(ws):
    s = logsumexp(ws)
    return np.exp(ws - s)


from typing import Union, Optional, Dict, List
import scipy.sparse as ss
import numpy as np
import functools

from snapatac2._snapatac2 import AnnData, AnnDataSet

def _group_by(x, groups):
    idx = groups.argsort()
    groups = groups[idx]
    x = x[idx]
    u, indices = np.unique(groups, return_index=True)
    splits = np.split(np.arange(x.shape[0]), indices[1:])
    return dict((label, x[indices, :]) for (label, indices) in zip(u, splits))

def _normalize(x, size_factor = None):
    result = x / (x.sum() / 1000000.0)
    if size_factor is not None:
        result /= size_factor
    return result

def _get_sizes(regions):
    def size(x):
        x = x.split(':')[1].split("-")
        return int(x[1]) - int(x[0])
    return np.array(list(size(x) for x in regions))

def aggregate_X(
    adata: Union[AnnData, AnnDataSet],
    group_by: Optional[Union[str, List]] = None,
    normalize: Optional[str] = None,
    inplace: bool = True,
) -> Union[np.ndarray, Dict, None]:
    """
    Aggregate values in adata.X in a row-wise fashion.

    Aggregate values in adata.X in a row-wise fashion. This is used to compute
    RPKM or RPM values stratified by user-provided groupings.

    Parameters
    ----------
    adata
        The AnnData or AnnDataSet object.
    group_by
        Group information, as adata.obs[group_by].
    normalize
        normalization method: "RPM" or "RPKM".
    inplace
        if True, results are added to adata.var.

    Returns
    -------
        If inplace=False, return the result as numpy array.
    """
 
    def norm(x):
        if normalize is None:
            return x
        elif normalize == "RPKM":
            size_factor = _get_sizes(adata.var_names) / 1000.0
            return _normalize(x, size_factor)
        elif normalize == "RPM":
            return _normalize(x)
        else:
            raise NameError("Normalization method must be 'RPKM' or 'RPM'")

    if group_by is None:
        row_sum = functools.reduce(
            lambda a, b: a + b,
            (np.ravel(chunk.sum(axis=0)) for chunk in adata.X.chunked(1000)),
        )
        row_sum = norm(row_sum)

        if inplace:
            adata.var["aggregate_X"] = row_sum
        else:
            return row_sum
    else:
        if isinstance(group_by, list):
            groups = adata.obs[group_by]
            groups = [tuple(groups[i, :]) for i in range(groups.shape[0])]
            out = np.empty(len(groups), dtype=object)
            out[:] = groups
            groups = out
            '''
            groups = np.array(list(
                '+'.join(map(lambda x: str(x), list(groups[i, :]))) for i in range(groups.shape[0])
            ))
            '''
        else:
            groups = adata.obs[group_by]
        cur_row = 0
        result = {}
        for chunk in adata.X.chunked(2000):
            n = chunk.shape[0]
            labels = groups[cur_row:cur_row+n]
            for key, mat in _group_by(chunk, labels).items():
                s = np.ravel(mat.sum(axis = 0))
                if key in result:
                    result[key] += s
                else:
                    result[key] = s
            cur_row = cur_row + n
        for k, v in result.items():
            result[k] = norm(v)
        if inplace:
            var = adata.var[:]
            for k, v in result.items():
                var[k] = v
            adata.var = var
        else:
            return result


""" Implementation of the scrublet algorithm for single-cell ATAC-seq data
"""
from tokenize import Name
import numpy as np
import scipy.sparse as ss
from typing import Optional, Union, Type, Tuple
from sklearn.neighbors import NearestNeighbors

from .._utils import get_binarized_matrix, chunks
from snapatac2._snapatac2 import AnnData
from snapatac2.tools._spectral import Spectral

def scrublet(
    adata: AnnData,
    features: Optional[Union[str, np.ndarray]] = "selected",
    n_comps: int = 15,
    sim_doublet_ratio: float = 2.0,
    expected_doublet_rate: float = 0.1,
    n_neighbors: Optional[int] = None,
    use_approx_neighbors=True,
    random_state: int = 0,
) -> None:
    """
    Compute probability of being a doublet using the scrublet algorithm.

    Parameters
    ----------
    adata
        AnnData object
    features
        Boolean index mask. True means that the feature is kept.
        False means the feature is removed.
    n_comps
        Number of PCs
    sim_doublet_ratio
        Number of doublets to simulate relative to the number of observed cells.
    expected_doublet_rate
    n_neighbors
        Number of neighbors used to construct the KNN graph of observed
        cells and simulated doublets. If `None`, this is 
        set to round(0.5 * sqrt(n_cells))
    use_approx_neighbors
        Whether to use 
    random_state
        Random state
    
    Returns
    -------
    It updates adata with the following fields:
    - ``adata.obs["doublet_score"]``: scrublet doublet score
    - ``adata.uns["scrublet"]["sim_doublet_score"]``: doublet scores of simulated doublets 
    """
    #- adata.uns["scrublet_cell_embedding"]: embedding of cells
    #- adata.uns["scrublet_sim_doublet_embedding"]: embedding of simulated doublets

    if isinstance(features, str):
        if features in adata.var:
            features = adata.var[features]
        else:
            raise NameError("Please call `select_features` first or explicitly set `features = None`")

    if features is None:
        count_matrix = adata.X[...]
    else:
        count_matrix = adata.X[:, features]

    if min(count_matrix.shape) == 0: raise NameError("Matrix is empty")

    if n_neighbors is None:
        n_neighbors = int(round(0.5 * np.sqrt(count_matrix.shape[0])))

    (doublet_scores_obs, doublet_scores_sim, manifold_obs, manifold_sim) = scrub_doublets_core(
        count_matrix, n_neighbors, sim_doublet_ratio, expected_doublet_rate,
        n_comps=n_comps,
        random_state=random_state
        )
    adata.obs["doublet_score"] = doublet_scores_obs
    adata.uns["scrublet_sim_doublet_score"] = doublet_scores_sim

def call_doublets(
    adata: AnnData,
    threshold: Union[str, float] = "gmm",
    random_state: int = 0,
    inplace: bool = True,
) -> Optional[Tuple[np.ndarray, float]]:
    """
    Find doublet score threshold for calling doublets.

    Need to run scrublet first.

    Parameters
    ----------
    adata
        AnnData object
    threshold
        Mannually specify a threshold or use one of the default methods to
        automatically identify a threshold:
        - "gmm": fit a 2-component gaussian mixture model
        - "scrublet": the method used in the scrublet paper
    random_state
        Random state
    inplace
        Whether update the AnnData object inplace
    
    Returns
    -------
    if ``inplace = True``, it updates adata with the following fields:
    - ``adata.obs["is_doublet"]``: boolean mask
    - ``adata.uns["scrublet"]["threshold"]``: saved threshold
    """

    if 'scrublet_sim_doublet_score' not in adata.uns:
        raise NameError("Please call `scrublet` first")

    doublet_scores_sim = adata.uns["scrublet_sim_doublet_score"]
    doublet_scores_obs = adata.obs["doublet_score"]

    if isinstance(threshold, float):
        thres = threshold
    elif threshold == "gmm":
        thres, _ = get_doublet_probability(
            doublet_scores_sim, doublet_scores_obs, random_state
        )
    elif threshold == "scrublet":
        from skimage.filters import threshold_minimum
        thres = threshold_minimum(doublet_scores_sim)
    else:
        raise NameError("Invalid value for the `threshold` argument")

    doublets = doublet_scores_obs >= thres
    if inplace:
        adata.uns["scrublet_threshold"] = thres
        adata.obs["is_doublet"] = doublets
    else:
        return (doublets, thres)

def scrub_doublets_core(
    count_matrix: ss.spmatrix,
    n_neighbors: int,
    sim_doublet_ratio: float,
    expected_doublet_rate: float,
    synthetic_doublet_umi_subsampling: float =1.0,
    n_comps: int = 30,
    random_state: int = 0,
) -> None:
    """
    Modified scrublet pipeline for single-cell ATAC-seq data.

    Automatically sets a threshold for calling doublets, but it's best to check 
    this by running plot_histogram() afterwards and adjusting threshold 
    with call_doublets(threshold=new_threshold) if necessary.

    Arguments
    ---------
    synthetic_doublet_umi_subsampling
        Rate for sampling UMIs when creating synthetic doublets. If 1.0, 
        each doublet is created by simply adding the UMIs from two randomly 
        sampled observed transcriptomes. For values less than 1, the 
        UMI counts are added and then randomly sampled at the specified
        rate.
    n_comps
        Number of principal components used to embed the transcriptomes prior
        to k-nearest-neighbor graph construction.

    Returns
    -------
    doublet_scores_obs_, doublet_errors_obs_,
    doublet_scores_sim_, doublet_errors_sim_,
    predicted_doublets_, z_scores_ 
    threshold_, detected_doublet_rate_,
    detectable_doublet_fraction_, overall_doublet_rate_,
    doublet_parents_, doublet_neighbor_parents_ 
    """
    total_counts_obs = count_matrix.sum(1).A.squeeze()

    print('Simulating doublets...')
    (count_matrix_sim, total_counts_sim, _) = simulate_doublets(
        count_matrix, total_counts_obs, sim_doublet_ratio,
        synthetic_doublet_umi_subsampling, random_state
    )

    E_obs_norm = get_binarized_matrix(count_matrix)
    E_sim_norm = get_binarized_matrix(count_matrix_sim)

    print('Spectral embedding ...')
    (manifold_obs, manifold_sim) = get_manifold(E_obs_norm, E_sim_norm, n_comps=n_comps)

    print('Calculating doublet scores...')
    (doublet_scores_obs, doublet_scores_sim) = calculate_doublet_scores(
        manifold_obs, manifold_sim, k = n_neighbors,
        exp_doub_rate = expected_doublet_rate,
        random_state = random_state,
    )

    return (doublet_scores_obs, doublet_scores_sim, manifold_obs, manifold_sim)

def get_manifold(obs_norm, sim_norm, n_comps=30, random_state=0):
    model = Spectral(n_dim=n_comps, distance="jaccard").fit(obs_norm, verbose=0)
    for c in chunks(obs_norm, 2000):
        model.extend(c)
    for c in chunks(sim_norm, 2000):
        model.extend(c)
    manifold = np.asanyarray(model.transform()[1])
    n = obs_norm.shape[0]
    manifold_obs = manifold[0:n, ]
    manifold_sim = manifold[n:, ]
    return (manifold_obs, manifold_sim)

def simulate_doublets(
    count_matrix: ss.spmatrix,
    total_counts: np.ndarray,
    sim_doublet_ratio: int = 2,
    synthetic_doublet_umi_subsampling: float = 1.0,
    random_state: int = 0,
) -> Tuple[ss.spmatrix, np.ndarray, np.ndarray]:
    """
    Simulate doublets by adding the counts of random cell pairs.

    Parameters
    ----------
    count_matrix
    total_counts
        Total insertion counts in each cell
    sim_doublet_ratio
        Number of doublets to simulate relative to the number of cells.
    synthetic_doublet_umi_subsampling
        Rate for sampling UMIs when creating synthetic doublets. If 1.0, 
        each doublet is created by simply adding the UMIs from two randomly 
        sampled observed transcriptomes. For values less than 1, the 
        UMI counts are added and then randomly sampled at the specified rate.

    Returns
    -------
    count_matrix_sim

    """
    n_obs = count_matrix.shape[0]
    n_sim = int(n_obs * sim_doublet_ratio)

    np.random.seed(random_state)
    pair_ix = np.random.randint(0, n_obs, size=(n_sim, 2))

    count_matrix_sim = count_matrix[pair_ix[:,0],:] + count_matrix[pair_ix[:,1],:]
    total_counts_sim = total_counts[pair_ix[:,0]] + total_counts[pair_ix[:,1]]

    if synthetic_doublet_umi_subsampling < 1:
        pass
        #count_matrix_sim, total_counts_sim = subsample_counts(
        #    count_matrix_sim, synthetic_doublet_umi_subsampling, total_counts_sim,
        #    random_seed=random_state
        #)
    return (count_matrix_sim, total_counts_sim, pair_ix)

def calculate_doublet_scores(
    manifold_obs: np.ndarray,
    manifold_sim: np.ndarray,
    k: int = 40,
    exp_doub_rate: float = 0.1,
    stdev_doub_rate: float = 0.03,
    random_state: int = 0,
) -> None:
    """
    Parameters
    ----------
    manifold_obs
        Manifold of observations
    manifold_sim
        Manifold of simulated doublets
    k
        Number of nearest neighbors
    exp_doub_rate
    stdev_doub_rate
    random_state
    """
    n_obs = manifold_obs.shape[0]
    n_sim = manifold_sim.shape[0]

    manifold = np.vstack((manifold_obs, manifold_sim))
    doub_labels = np.concatenate(
        (np.zeros(n_obs, dtype=int), np.ones(n_sim, dtype=int))
    )

    # Adjust k (number of nearest neighbors) based on the ratio of simulated to observed cells
    k_adj = int(round(k * (1 + n_sim / float(n_obs))))
    
    # Find k_adj nearest neighbors
    neighbors = NearestNeighbors(
        n_neighbors=k_adj, metric="euclidean"
        ).fit(manifold).kneighbors(return_distance=False)
    
    # Calculate doublet score based on ratio of simulated cell neighbors vs. observed cell neighbors
    doub_neigh_mask = doub_labels[neighbors] == 1
    n_sim_neigh = doub_neigh_mask.sum(1)
    n_obs_neigh = doub_neigh_mask.shape[1] - n_sim_neigh
    
    rho = exp_doub_rate
    r = n_sim / float(n_obs)
    nd = n_sim_neigh.astype(float)
    ns = n_obs_neigh.astype(float)
    N = float(k_adj)
    
    # Bayesian
    q=(nd+1)/(N+2)
    Ld = q*rho/r/(1-rho-q*(1-rho-rho/r))

    se_q = np.sqrt(q*(1-q)/(N+3))
    se_rho = stdev_doub_rate

    se_Ld = q*rho/r / (1-rho-q*(1-rho-rho/r))**2 * np.sqrt((se_q/q*(1-rho))**2 + (se_rho/rho*(1-q))**2)

    doublet_scores_obs = Ld[doub_labels == 0]
    doublet_scores_sim = Ld[doub_labels == 1]
    doublet_errors_obs = se_Ld[doub_labels==0]
    doublet_errors_sim = se_Ld[doub_labels==1]

    return (doublet_scores_obs, doublet_scores_sim)

def get_doublet_probability(
    doublet_scores_sim: np.ndarray,
    doublet_scores: np.ndarray,
    random_state: int = 0,
):
    from sklearn.mixture import BayesianGaussianMixture

    X = doublet_scores_sim.reshape((-1, 1))
    gmm = BayesianGaussianMixture(
        n_components=2, n_init=10, max_iter=1000, random_state=random_state
    ).fit(X)
    i = np.argmax(gmm.means_)

    probs_sim = gmm.predict_proba(X)[:,i]
    vals = X[probs_sim > 0.5]
    threshold = X.max() if vals.size == 0 else vals.min()

    return (threshold, gmm.predict_proba(doublet_scores.reshape((-1, 1)))[:,i])



from pathlib import Path
import numpy as np
from typing import Optional, Sequence, Union, Set, Mapping

from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as internal

def export_bed(
    adata: Union[AnnData, AnnDataSet],
    group_by: Union[str, Sequence[str]],
    selections: Optional[Set[str]] = None,
    ids: Optional[Union[str, np.ndarray]] = None,
    out_dir: Path = "./",
    prefix: str = "",
    suffix: str = ".bed.gz",
) -> Mapping[str, str]:
    """
    Export base-resolution TN5 insertion sites as BED format file.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    group_by
        Group the cells. If a `str`, groups are obtained from
        `.obs[group_by]`.
    selections
        Export only the selected groups.
    ids
        Cell ids add to the bed records, default to the `.obs_names`.
    out_dir
        Directory for saving the outputs.
    prefix
        Text that adds to the output file name.
    suffix
        Text that adds to the output file name.

    Returns
    -------
    A dictionary contains `(groupname, filename)` pairs. The file names are
    formatted as `{prefix}{groupname}{suffix}`.
    """
    if isinstance(group_by, str):
        group_by = adata.obs[group_by].astype("str")
    
    if ids is None:
        ids = adata.obs[:, 0].astype("str")
    elif isinstance(ids, str):
        ids = adata.obs[ids].astype("str")

    return internal.export_bed(
        adata, list(ids), list(group_by), selections, str(out_dir), prefix, suffix,
    )


from pathlib import Path
from typing import Union
import pandas as pd
from snapatac2._snapatac2 import AnnData, read_mtx

def read_10x_mtx(
    path: Union[Path, str],
    prefix: str = None,
) -> AnnData:
    """
    Read 10x-Genomics-formatted mtx directory.

    Parameters
    ----------
    path
        Path to directory for `.mtx` and `.tsv` files,
    prefix
        Any prefix before `matrix.mtx`, `genes.tsv` and `barcodes.tsv`. For instance,
        if the files are named `patientA_matrix.mtx`, `patientA_genes.tsv` and
        `patientA_barcodes.tsv` the prefix is `patientA_`.
    Returns
    -------
    An :class:`~anndata.AnnData` object
    """

    def get_files(prefix, names):
        return list(filter(
            lambda x: x.is_file(),
            map(lambda x: Path(path + "/" + prefix + x), names)
        ))

    prefix = "" if prefix is None else prefix

    matrix_files = get_files(prefix, ["matrix.mtx", "matrix.mtx.gz"])
    if len(matrix_files) == 1:
        adata = read_mtx(matrix_files[0]).T  # transpose the data
    else:
        raise ValueError("Expecting a single 'matrix.mtx' or 'matrix.mtx.gz' file")

    feature_files = get_files(
        prefix,
        ["genes.tsv", "genes.tsv.gz", "features.tsv", "features.tsv.gz"]
    )
    if len(feature_files) == 1:
        adata.var['regions'] = pd.read_csv(feature_files[0], header=None, sep='\t')[0].values
    else:
        raise ValueError("Expecting a single feature file")

    barcode_files = get_files(prefix, ["barcodes.tsv", "barcodes.tsv.gz"])
    if len(barcode_files) == 1:
        adata.obs_names = pd.read_csv(barcode_files[0], header=None)[0].values
    else:
        raise ValueError("Expecting a single barcode file")

    return adata